use std::os;

use core::Source;
use sources::PathSource;
use ops;
use util::{CargoResult, ProcessError};

pub struct TestOptions<'a> {
    pub compile_opts: ops::CompileOptions<'a>,
    pub no_run: bool,
    pub name: Option<&'a str>,
}

pub fn run_tests(manifest_path: &Path,
                 options: &mut TestOptions,
                 test_args: &[String]) -> CargoResult<Option<ProcessError>> {
    let mut source = try!(PathSource::for_path(&manifest_path.dir_path()));
    try!(source.update());

    let mut compile = try!(ops::compile(manifest_path, &mut options.compile_opts));
    if options.no_run { return Ok(None) }
    compile.tests.sort();

    let target_name = options.name;
    let mut tests_to_run = compile.tests.iter().filter(|&&(ref test_name, _)| {
        target_name.map_or(true, |target_name| target_name == test_name.as_slice())
    });

    let cwd = try!(os::getcwd());
    for &(_, ref exe) in tests_to_run {
        let to_display = match exe.path_relative_from(&cwd) {
            Some(path) => path,
            None => exe.clone(),
        };
        let cmd = try!(compile.process(exe, &compile.package)).args(test_args);
        try!(options.compile_opts.shell.concise(|shell| {
            shell.status("Running", to_display.display().to_string())
        }));
        try!(options.compile_opts.shell.verbose(|shell| {
            shell.status("Running", cmd.to_string())
        }));
        match cmd.exec() {
            Ok(()) => {}
            Err(e) => return Ok(Some(e))
        }
    }

    if options.name.is_some() { return Ok(None) }

    if options.compile_opts.env == "bench" { return Ok(None) }

    let mut libs = compile.package.get_targets().iter().filter_map(|target| {
        if !target.get_profile().is_doctest() || !target.is_lib() {
            return None
        }
        Some((target.get_src_path(), target.get_name()))
    });

    for (lib, name) in libs {
        try!(options.compile_opts.shell.status("Doc-tests", name));
        let mut p = try!(compile.process("rustdoc", &compile.package))
                           .arg("--test").arg(lib)
                           .arg("--crate-name").arg(name)
                           .arg("-L").arg(&compile.root_output)
                           .arg("-L").arg(&compile.deps_output)
                           .cwd(compile.package.get_root());

        // FIXME(rust-lang/rust#16272): this should just always be passed.
        if test_args.len() > 0 {
            p = p.arg("--test-args").arg(test_args.connect(" "));
        }

        for (pkg, libs) in compile.libraries.iter() {
            for lib in libs.iter() {
                let mut arg = pkg.get_name().as_bytes().to_vec();
                arg.push(b'=');
                arg.push_all(lib.as_vec());
                p = p.arg("--extern").arg(arg.as_slice());
            }
        }

        try!(options.compile_opts.shell.verbose(|shell| {
            shell.status("Running", p.to_string())
        }));
        match p.exec() {
            Ok(()) => {}
            Err(e) => return Ok(Some(e)),
        }
    }

    Ok(None)
}

pub fn run_benches(manifest_path: &Path,
                   options: &mut TestOptions,
                   args: &[String]) -> CargoResult<Option<ProcessError>> {
    let mut args = args.to_vec();
    args.push("--bench".to_string());

    run_tests(manifest_path, options, args.as_slice())
}
